-----------------------------------------------------------------------------------------------
-- Client Lua Script for StatView
-- Copyright Celess. Portions (C) Foxykeep via DPSStats. All rights reserved.
-----------------------------------------------------------------------------------------------

require "Window"


-----------------------------------------------------------------------------------------------
-- StatView Module Definition
-----------------------------------------------------------------------------------------------
local _Version,_VersionMinor = 70923,0
local _AddonName,_VersionWindow = "StatView", 1
local StatView = {}

-----------------------------------------------------------------------------------------------
-- Constants
-----------------------------------------------------------------------------------------------

local _eSaveLevel = GameLib.CodeEnumAddonSaveLevel

local _defaultSettings = {	--type: 0 custom, 1 single check, 2 slider, 3 color, 4 radio, 5 text, 6 number, 7 check parts, 8 signal, 9 drop, 10 check drop
	-- Base
	arPanels = {
		{								-- main
			--bSizeChanged = false,
			--wndPosition = nil,
			--wndAnchors = nil,
			--wscPosition = nil,
		},
	},
	isVisible = true,

	-- Other

	sSettingsName = "Default",
	nSettingsVersion = 1,
}


local _eClass = GameLib.CodeEnumClass

local _eRole = {
	DPS = 1,
	Heal = 2,
	Tank = 3,
}

local _eDPS  = _eRole.DPS
local _eHeal = _eRole.Heal
local _eTank = _eRole.Tank

local _arLabelNames = { "LabelAP", "LabelST", "LabelCH", "LabelCHS", "LabelMH", "LabelMHS", "LabelVI", "LabelPI", "LabelGL", "LabelGLM", "LabelCHM", }
local _arValueNames = { "AP", "ST", "CH", "CHS", "MH", "MHS", "VI", "PI", "GL", "GLM", "CHM", }		-- name keys
local _tDR = {																						-- diminishing returns threshholds
	[_eDPS] =  { nil, 30, 30, 250, 60, 100, 30, 100, },		-- 7:Vigor, 8:Pierce
	[_eHeal] = { nil, nil, 30, 250, 60, 100, 1.5, 30, },	-- 7:Focus Rec, 8:Intensity, 2:Focus Pool
	[_eTank] = { nil, 30, 30, 250, 60, 100, 0, 30, },		-- 7:Mitigation, 8:Deflect
}
local _tRoleRows = {																				-- diminishing returns threshholds
	[_eDPS] = #_arValueNames - 3,
	[_eHeal] = #_arValueNames - 3,
	[_eTank] = #_arValueNames,
}

local _nWindowMargin = 33		-- margin (14 * 2) + between label and value (12) - user squeeze if wants (7)
local _nExpandedHeight = 60

local _crTextNormal = "UI_WindowTextDefault"
local _crTextDR = "ConTough"							--ffff8000
local _crTextDR_Light = "ffffA050"
local _crTextChange = ApolloColor.new("ChatSay")		--"cyan"

local _tHealClass = {
	[_eClass.Esper] = true,
	[_eClass.Medic] = true,
	[_eClass.Spellslinger] = true,
}

local _LAttributeAssaultPower = Apollo.GetString("AttributeAssaultPower")
local _LAttributeSupportPower = Apollo.GetString("AttributeSupportPower")
local _LAttributeStrikethrough = Item.GetPropertyName(Unit.CodeEnumProperties.BaseAvoidReduceChance)
local _LAttributeFocusPool = Item.GetPropertyName(Unit.CodeEnumProperties.BaseFocusPool)
local _LAttributeMultiHitChance = Item.GetPropertyName(Unit.CodeEnumProperties.BaseMultiHitChance)
local _LAttributeMultiHitSeverity = Item.GetPropertyName(Unit.CodeEnumProperties.BaseMultiHitAmount)
local _LAttributeVigor = Item.GetPropertyName(Unit.CodeEnumProperties.BaseVigor)
local _LAttributeFocusRecovery = Item.GetPropertyName(Unit.CodeEnumProperties.BaseFocusRecoveryInCombat)
local _LAttributeMitigation = Item.GetPropertyName(Unit.CodeEnumProperties.DamageMitigationPctOffsetPhysical)
local _LAttributePierce = Item.GetPropertyName(Unit.CodeEnumProperties.IgnoreArmorBase)
local _LAttributeIntensity = Item.GetPropertyName(Unit.CodeEnumProperties.BaseIntensity)
local _LAttributeDeflect = Item.GetPropertyName(Unit.CodeEnumProperties.BaseAvoidChance)
local _LAttributeGlance = Item.GetPropertyName(Unit.CodeEnumProperties.BaseGlanceChance)
local _LAttributeGlanceMitigation = Item.GetPropertyName(Unit.CodeEnumProperties.BaseGlanceAmount)
local _LAttributeCritHitMitigation = Item.GetPropertyName(Unit.CodeEnumProperties.BaseCriticalMitigation)

local _floor = math.floor
local _abs = math.abs
local _format = string.format
local _len = string.len
local _tsort = table.sort
local _byte = string.byte
local _char = string.char

local function _dummy() end
local print = _dummy

local function _print(...)
	local s = ""
	for i = 1, select("#",...) do
		s = s .. tostring(select(i,...)) .. " "
	end
	Print(s)
end

local function DeepCopy(o)
	if type(o) ~= 'table' then return o end

	local copy = {}
	for k,v in next, o do copy[k] = DeepCopy(v) end
	return copy
end

local function SettingsCopy(o)
	local copy = {}
	for k,v in next, o do
		if type(v) == "table" and v.nControlType then v = v.default end
		copy[k] = DeepCopy(v)
	end
	return copy
end

local function PostChat(s)
	ChatSystemLib.PostOnChannel(ChatSystemLib.ChatChannel_System, s)
end


-----------------------------------------------------------------------------------------------
-- Initialization
-----------------------------------------------------------------------------------------------

function StatView:new(o)
	o = self
	--o = o or {}
	--setmetatable(o, self)
	--self.__index = self

	o.L = o.L or _L or {}
	setmetatable(o.L, { __index = function(t,k) return tostring(k) end })

	o._AddonName = _AddonName
	o._defaultSettings = _defaultSettings
	o._resetSettings = _resetSettings
	o._print = _print

	o.arPanels = { }

	o.statsLast = { }													-- last stats
	o.stats = { }														-- current stats
	o.colors = { }														-- current colors

	o.wLabels = { }														-- label window handles
	o.wValues = { }														-- value window handles

	return o
end

function StatView:Init()
	local bHasConfigureFunction = false
	local strConfigureButtonText = ""
	local tDependencies = { }
	Apollo.RegisterAddon(self, bHasConfigureFunction, strConfigureButtonText, tDependencies)
end


-----------------------------------------------------------------------------------------------
-- StatView OnLoad
-----------------------------------------------------------------------------------------------

function StatView:OnLoad()
	self.onXmlDocLoadedCalled = false

	self.userSettings = SettingsCopy(_defaultSettings)

	-- create form
	self.xmlDoc = XmlDoc.CreateFromFile("StatView.xml")
	self.xmlDoc:RegisterCallback("OnDocLoaded", self)

	Apollo.RegisterEventHandler("StatView_ShowMenu", "OnShowMenu", self)
	Apollo.RegisterEventHandler("InterfaceMenuListHasLoaded", "OnInterfaceMenuListHasLoaded", self)

	Apollo.RegisterSlashCommand("sv", "OnCommand", self)
	Apollo.RegisterSlashCommand("SV", "OnCommand", self)
	Apollo.RegisterSlashCommand("StatView", "OnCommand", self)
	Apollo.RegisterSlashCommand("statview", "OnCommand", self)
end

function StatView:OnInterfaceMenuListHasLoaded()
	Event_FireGenericEvent("InterfaceMenuList_NewAddOn", _AddonName,
		{ "StatView_ShowMenu", "", "IconSprites:Icon_Windows_UI_CRB_Checkmark" })
end


-----------------------------------------------------------------------------------------------
-- StatMeter OnDocLoaded
-----------------------------------------------------------------------------------------------

function StatView:OnDocLoaded()
	if self.xmlDoc == nil or not self.xmlDoc:IsLoaded() then
		return
	end

	Event_FireGenericEvent("OneVersion_ReportAddonInfo", _AddonName, 1, _Version, _VersionMinor)

	-- Load the window

	self.wndMain = Apollo.LoadForm(self.xmlDoc, _AddonName, nil, self)
	if not self.wndMain then
		Apollo.AddAddonErrorText(self, "Could not load the main window for some reason.")
		return
	end

	Apollo.RegisterEventHandler("FrameCount",					"OnFrame", self)

	Apollo.RegisterEventHandler("ResolutionChanged",			"OnResolutionChanged", self)

	self.tUpdateTimer = ApolloTimer.Create(0.5, true, "OnUpdateTimer", self)

	self.onXmlDocLoadedCalled = true
	self:InitializeWindows()
end


-----------------------------------------------------------------------------------------------
-- StatView Functions
-----------------------------------------------------------------------------------------------

function StatView:OnFrame()
	local o,w,panel,player = self.userSettings, self.wndMain, self.tPanel, self.unitPlayer
	if not w or not o.isVisible or not player then
		return
	end

	-- update stats
	local ap,sp = player:GetAssaultPower(), player:GetSupportPower()
	local bIsAP = ap >= sp
	local eRole = bIsAP and _eDPS or (self.bIsHealClass and _eHeal or _eTank)
	local dr,colors = _tDR[eRole], self.colors
	local bDps,bHeal = eRole == _eDPS, eRole == _eHeal
	local bTank = not (bDps or bHeal)

	local stats, statsLast = self.statsLast, self.stats
	self.statsLast, self.stats = statsLast, stats				-- swap tables for later

	stats[1] = bIsAP and ap or sp
	stats[2] = bDps and player:GetStrikethroughChance().nAmount or (bHeal and player:GetMaxFocus() or player:GetStrikethroughChance().nAmount)
	stats[3] = player:GetCritChance().nAmount
	stats[4] = player:GetCritSeverity().nAmount
	stats[5] = player:GetMultiHitChance().nAmount
	stats[6] = player:GetMultiHitAmount().nAmount
	stats[7] = bDps and player:GetVigor().nAmount or (bHeal and player:GetFocusRegenInCombat().nAmount or player:GetPhysicalMitigation().nAmount)
	stats[8] = bDps and player:GetArmorPierce().nAmount or (bHeal and player:GetIntensity().nAmount or player:GetDeflectChance().nAmount)
	if bTank then
	stats[9] = player:GetGlanceChance().nAmount
	stats[10] = player:GetGlanceAmount().nAmount
	stats[11] = player:GetCriticalMitigation().nAmount
	end

	-- roll though updates
	local wLabels, wValues = self.wLabels, self.wValues
	for i,v in next, stats do
		local vOld,value,dr = statsLast[i], wValues[i], dr[i]

		if v ~= vOld then
			value:SetText(v and Apollo.FormatNumber(v, 2, true) or "")		--FIXME: needs to use the short format guard from TargetView to skip unnessary C calls
		end

		local color = dr and v and v >= dr and _crTextDR or (v ~= vOld and _crTextChange or _crTextNormal)
		if color ~= colors[i] then
			value:SetTextColor(color)
			colors[i] = color
		end
	end

	if bIsAP ~= self.bIsAP then
		wLabels[1]:SetText(bIsAP and _LAttributeAssaultPower or _LAttributeSupportPower)
		self.bIsAP = bIsAP
	end

	if eRole ~= self.eRole then
		wLabels[2]:SetText(bDps and _LAttributeStrikethrough or (bHeal and _LAttributeFocusPool or _LAttributeStrikethrough))
		wLabels[7]:SetText(bDps and _LAttributeVigor or (bHeal and _LAttributeFocusRecovery or _LAttributeMitigation))
		wLabels[8]:SetText(bDps and _LAttributePierce or (bHeal and _LAttributeIntensity or _LAttributeDeflect))

		local p = panel._pos
		local l,t,r,b = w:GetAnchorOffsets()
		w:SetAnchorOffsets(l, t, r, t + panel._height + (bTank and _nExpandedHeight or 0))
		wLabels[9]:Show(bTank)
		wLabels[10]:Show(bTank)
		wLabels[11]:Show(bTank)
		wValues[9]:Show(bTank)
		wValues[10]:Show(bTank)
		wValues[11]:Show(bTank)

		self.eRole = eRole
	end

end


-- create destroy		-----------------------------------------------------------------------

function StatView:UpdateLayout()
	local ipa = self.arPanels

	for _,v in ipairs(ipa) do
		self:UpdatePanelLayout(v)
	end
end

function StatView:UpdatePanelLayout(panel)
	local w = panel.w
	local bSizeChanged = panel.bSizeChanged
	if self.bInLayout or not panel.bShow or not w then
		return
	end
	self.bInLayout = true

	local left,top,right,bottom = w:GetAnchorOffsets()
	local _right,_bottom = right, bottom

	local widthNow, heightNow = _abs(left - right), _abs(bottom - top)
	local width, height = bSizeChanged and widthNow or panel._width,		-- init width and height to default
		bSizeChanged and heightNow or panel._height

	if panel.scroll then
		panel.scroll:ArrangeChildrenVert()
	end

	-- measure
	local rows = 0
	local nLabelWidth, nValueWidth = 0, 0

	local labels,values = panel.labels, panel.values
	for i,v in next, labels do
		local textWidth = Apollo.GetTextWidth("Default", v:GetText())
			+ Apollo.GetTextWidth("Default", values[i]:GetText()) + _nWindowMargin

		if textWidth > width then
			width = textWidth
		end
	end

	-- set
	if width ~= widthNow then
		right = right + (width - widthNow)
	end
	if height ~= heightNow then
		bottom = bottom + (height - heightNow)
	end
	if right ~= _right or bottom ~= _bottom or bMustSet then
		w:SetAnchorOffsets(left, top, right, bottom)
	end

	panel.nLastWidth = width
	panel.nLastHeight = height
	self.bInLayout = nil
end


-- visibility timer		-----------------------------------------------------------------------

function StatView:OnUpdateTimer()
	self:OnUpdateVisibility()
end

function StatView:OnUpdateVisibility()
	self:UpdateVisibilityCommon()

	self:UpdatePlayer()
	if self.bRequestLayout then
		self:UpdateLayout()
		self.bRequestLayout = nil
	end
	self:UpdateVisibilityExternal()
end

function StatView:UpdateVisibilityCommon()
	local o,w,player = self.userSettings, self.wndMain, self.unitPlayer

	if not player or not player:IsValid() then
		player = GameLib.GetPlayerUnit()
		self.unitPlayer = player or self.unitPlayer

		local eClass = player and player:GetClassId()
		self.bIsHealClass = eClass and _tHealClass[eClass]
		self.eClass = eClass or self.eClass
	end

	local nUiScale = Apollo.GetConsoleVariable("ui.Scale")
	if nUiScale ~= self.nUiScale and #self.arPanels then
		self.bRequestLayout = true
		self.nUiScale = nUiScale
	end

	if self._log and self.nAge > 3 then														--PROF:
		for i,v in ipairs(self._log) do														--PROF:
			if type(v) == "table" then print(unpack(v)) else print(v) end					--PROF:
		end																					--PROF:
		self._log = nil																		--PROF:
	end																						--PROF:
end

function StatView:UpdateVisibilityExternal()
	-- external mini buttons etc
end

function StatView:UpdatePlayer()
end


-- other system events	-----------------------------------------------------------------------

function StatView:OnResolutionChanged()
	self.nUiScale = nil													-- force scale re-evaluation
	self:OnUpdateVisibility()											-- immediate, so no post rez change pop
end


-----------------------------------------------------------------------------------------------
-- StatView Form Functions
-----------------------------------------------------------------------------------------------

function StatView:OnWindowMove(wndHandler, wndControl)
	local o,ipa = self.userSettings, self.arPanels
	if wndHandler ~= wndControl or self.bInLayout then
		return
	end
	for i,v in next, ipa do
		if v.w == wndControl then
			local left,top,right,bottom = wndControl:GetAnchorOffsets()
			left,top = _abs(right - left), _abs(bottom - top)
			if v.nLastWidth ~= left or v.nLastHeight ~= top then
				v.bSizeChanged = true
			end
			if not o.bRelative then
				self:OnPositionRelativeChanged(v, true)
			end
		end
	end
	self:UpdateLayout()
end


-----------------------------------------------------------------------------------------------
-- OptionsForm Form Functions
-----------------------------------------------------------------------------------------------

function StatView:OnEnablerButtonChecked()
end

function StatView:OnPositionRelativeChanged(panel, bFreeMove)
	local o,w = self.userSettings, panel.w

	local p0,p1,p2,p3 = w:GetAnchorOffsets()
	local a0,a1,a2,a3 = w:GetAnchorPoints()
	local e = Apollo.GetDisplaySize()
	local u,v = e.nWidth, e.nHeight

	p0,p1,p2,p3 = p0 + u * a0, p1 + v * a1, p2 + u * a2, p3 + v * a3			-- normalize position to anchors 0,0,0,0

	if o.bRelative then
		a0 = o.bRelative and (o.bAnchorRight and 1 or 0) or a0
		a1 = o.bRelative and (o.bAnchorBottom and 1 or 0) or a1
	else
		local r0,r1,r2,r3 = p0 / u, p1 / v, p2 / u, p3 / v						-- get extent ratios to edges
		a0 = r0 < .3 and 0 or r2 > .7 and 1 or .5				-- left right
		a1 = r1 < .25 and 0 or r3 > .7 and 1 or .5				-- top bottom
	end
	a2,a3 = a0, a1

	if not bFreeMove then
		if p0 < 0 then p0,p2 = 0,				p2 - p0 end		-- left			-- clip position extents for each edge, do this on load
		if p1 < 0 then p1,p3 = 0,				p3 - p1 end		-- top
		if p2 > u then p0,p2 = u - (p2 - p0),	u		end		-- right
		if p3 > v then p1,p3 = v - (p3 - p1),	v		end		-- bottom
	end

	w:SetAnchorPoints(a0, a1, a2, a3)
	w:SetAnchorOffsets(p0 - u * a0, p1 - v * a1, p2 - u * a2, p3 - v * a3)		-- translate position to new anchors
end

function StatView:OnPositionOptionsChanged()
	local wo,ipa = self.wndOpts, self.arPanels
	for _,v in ipairs(ipa) do
		self:OnPositionRelativeChanged(v)
	end

	if wo and wo:IsValid() then
		self:OnEnablerButtonChecked()
	end
end


-----------------------------------------------------------------------------------------------
-- StatView Startup
-----------------------------------------------------------------------------------------------

function StatView:OnSave(eType)
	local tSave

	if eType == GameLib.CodeEnumAddonSaveLevel.Character then
		local o = self.userSettings

		local ipa = self.arPanels
		for i,v in next, o.arPanels do
			local panel = ipa[i]
			if panel.bSavePos then
				v.bSizeChanged = panel.bSizeChanged
				v.wndPosition = { panel.w:GetAnchorOffsets() }
				if panel.bSaveAnchors then
					v.wndAnchors = { panel.w:GetAnchorPoints() }
				end
				if panel.scroll then
					v.wscPosition = panel.scroll:GetVScrollPos()
				end
			end
		end

		tSave = { }
	end

	if eType == GameLib.CodeEnumAddonSaveLevel.Realm then
	end

	if tSave then
		local o = self.userSettings

		for k,v in pairs(o) do
			local t = _defaultSettings[k]
			if (type(t) == "table" and t.eSaveLevel or _eSaveLevel.Character) == eType then
				tSave[k] = DeepCopy(v)
			end
		end
	end

	return tSave
end

function StatView:OnRestore(eType, tSave)
	if tSave then
		local o,_ds = self.userSettings, _defaultSettings

		local nVersion = tSave["nSettingsVersion"] or 0

		for k,v in pairs(tSave) do
			local t = _defaultSettings[k]
			if (type(t) == "table" and t.eSaveLevel or _eSaveLevel.Character) == eType then
				-- migrations
				if nVersion < 1 and k == "wndPosition" then k,v = "arPanels", {{wndPosition=v}} end
				-- filters
				if k == "nSettingsVersion" then v = o.nSettingsVersion end
				o[k] = DeepCopy(v)
			end
			-- migrations
		end
	end

	if eType == GameLib.CodeEnumAddonSaveLevel.Realm then
	end

	if eType == GameLib.CodeEnumAddonSaveLevel.Character then
		self:InitializeWindows()
	end
end

function StatView:InitializeWindows()
	local o,w = self.userSettings, self.wndMain

	self:OnCurrentToCommand()

	self.nUiScale = self.nUiScale or Apollo.GetConsoleVariable("ui.Scale")

	-- rest needs docload
	if not self.onXmlDocLoadedCalled then
		return
	end

	-- panels
	local ipa = self.arPanels

	local tPanel = self.tPanel or { }; ipa[1] = tPanel
	tPanel.bShow = true									-- o.isVisible will control whole display UI
	tPanel.w = w
	tPanel.labels = self.wLabels
	tPanel.values = self.wValues
	
	-- default size and position
	local p = w and { w:GetAnchorOffsets() }
	tPanel._pos = p
	tPanel._anchors = w and { w:GetAnchorPoints() }
	tPanel._width = p and _abs(p[3] - p[1])
	tPanel._height = p and _abs(p[4] - p[2])
	tPanel.bSavePos = true
	tPanel.bSaveAnchors = true
	self.tPanel = tPanel

	-- collect window handles
	local wLabels, wValues = self.wLabels, self.wValues
	for i = 1, #_arValueNames, 1 do
		wLabels[i] = w:FindChild(_arLabelNames[i])
		wValues[i] = w:FindChild(_arValueNames[i])
	end
	
	-- set localized
	wLabels[5]:SetText(_LAttributeMultiHitChance)
	wLabels[6]:SetText(_LAttributeMultiHitSeverity)
	wLabels[9]:SetText(_LAttributeGlance)
	wLabels[10]:SetText(_LAttributeGlanceMitigation)
	wLabels[11]:SetText(_LAttributeCritHitMitigation)

	-- saved position
	for i,v in next, o.arPanels do
		local panel = ipa[i]
		if panel.bSavePos then
			panel.bSizeChanged = v.bSizeChanged
			panel.w:SetAnchorOffsets(unpack(v.wndPosition or panel._pos))
			if panel.bSaveAnchors then
				panel.w:SetAnchorPoints(unpack(v.wndAnchors or panel._anchors))
			end
			self.bRequestLayout = true
		end
	end

	self:OnPositionOptionsChanged()					-- ensure windows in bounds

	self:OnUpdateTimer()							-- layout

	for i,v in next, o.arPanels do					-- saved scroll bars, wont have anything to scroll to until layout does arrange vertical
		local panel = ipa[i]
		if panel.bSavePos and panel.scroll and v.wscPosition then
			panel.scroll:SetVScrollPos(v.wscPosition)
		end
	end

	w:SetOpacity(70 / 100)
	w:Show(o.isVisible)
end


-----------------------------------------------------------------------------------------------
-- StatView Commands
-----------------------------------------------------------------------------------------------

function StatView:OnShowMenu()				-- used by InterfaceMenuList
	self:OnCommand()
end

function StatView:OnCurrentToCommand()
	local o = self.userSettings
	print = o.bIsDebug and _print or _dummy
	self.print = print
end

function StatView:OnCommand(sCommand, sArgs)
	local o,w,L = self.userSettings, self.wndMain, self.L

	sArgs = sArgs and sArgs:lower() or ""
	local sArg1, sArg2 = sArgs:match("([^ ]+) *(.-) *$")	-- '  aaa   bb cc  ' to 'aaa', 'bb cc'

	if not sArg1 or sArg1 == "" then		-- toggle display window
		if _stats then _stats = nil end		-- clear history either way
		o.isVisible = not o.isVisible		-- toggle value
		w:Show(o.isVisible)					-- correct the window

	elseif sArg1 == "reset" then
		self.userSettings = SettingsCopy(_defaultSettings)
		self:InitializeWindows()

	elseif sArg1 == "debug" then
		o.bIsDebug = not o.bIsDebug
		self:OnCurrentToCommand()
		PostChat(L["debug"] .. " " .. (o.bIsDebug and L["on"] or L["off"]))

	elseif sArg1 == "help" then
		local _addonname = _AddonName:lower()
		PostChat(_format(L["%s commands:"],_AddonName))
		PostChat(_format(L[" /%s - show display"],_addonname))
		PostChat(_format(L[" /%s reset - reset to defaults"],_addonname))
		PostChat(_format(L[" /%s help - help message"],_addonname))
	end
end


-----------------------------------------------------------------------------------------------
-- StatView Instance
-----------------------------------------------------------------------------------------------

local StatViewInst = StatView:new()
StatViewInst:Init()

